﻿using System;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Text;

namespace FTPClient
{
    class TftpClient
    {
        public static Socket socket;
        public static IPEndPoint ipEndp;
        public static EndPoint endpoint;
        public static string status;
        private static int OpCode;
        private static int blockNumber = 1;
        public static NetworkStream ns;
        public static StreamReader sr;
        public static StreamWriter sw;

        private const string FILE_PATH = @"C:\temp\client";
        private static string filename;
        private const string MODE = "ascii";

        /// <summary>
        /// Enum for packet types
        /// </summary>
        enum PACKET_TYPE
        {
            RRQ = 1,
            WRQ,
            DATA,
            ACK,
            ERROR
        }

        enum ERROR_CODE
        {
            Not_defined,
            File_not_found,
            Access_violation,
            Disk_full,
            Illegal_TFTP_operation,
            Unknown_transfer_ID,
            File_already_exists,
            No_such_user
        }

        public string host { get; set; }
        public int port { get; set; }

        public TftpClient(string host, int port)
        {
            this.host = host;
            this.port = port;
        }

        /// <summary>
        /// Initialize connection and send a read request
        /// </summary>
        /// <param name="readFile">File to read</param>
        public void InitializeAndRead(string readFile)
        {
            Initialize();
            SendRRQ(readFile);
        }

        /// <summary>
        /// Initialize connection and send write request
        /// </summary>
        /// <param name="writeFile">File to write</param>
        public void InitializeAndWrite(string writeFile)
        {
            Initialize();
            SendWRQ(writeFile);
        }

        public void SendWRQ(string writeFile)
        {
            filename = writeFile;
            byte[] packet = MakeRequestPacket(PACKET_TYPE.WRQ);

            socket.SendTo(packet, endpoint);
            Console.WriteLine($"Sent Write Request for file: {writeFile}");
            if (!WaitAck()) return;
            HandleWriteRequest(writeFile);
        }

        /// <summary>
        /// Sends a Read Request with given file name
        /// </summary>
        public void SendRRQ(string readFile)
        {
            filename = readFile;
            byte[] packet = MakeRequestPacket(PACKET_TYPE.RRQ);

            socket.SendTo(packet, endpoint);
            Console.WriteLine($"Sent Read Request for file: {filename}");
            HandleReadRequest();
        }

        private bool WaitAck()
        {
            byte[] buffer = new byte[512];
            socket.ReceiveFrom(buffer, ref endpoint);
            int opcode = buffer[0] | buffer[1] << 8;
            if (opcode == (int)PACKET_TYPE.ACK) return true;
            if (opcode == (int)PACKET_TYPE.ERROR) ReportError(buffer);
            return false;
        }

        private byte[] MakeRequestPacket(PACKET_TYPE type)
        {
            byte[] packet = new byte[4 + filename.Length + MODE.Length];
            packet[0] = (byte)type;
            packet[1] = (byte)((int)type >> 8);
            OpCode = (int)type;

            int index = 2;
            // BUG : if filename is larger than 512 bytes causes problems
            foreach (char letter in filename)
            {
                packet[index] = (byte)letter;
                index += 1;
            }
            index += 1;

            foreach (char letter in MODE)
            {
                packet[index] = (byte)letter;
                index += 1;
            }
            return packet;
        }

        /// <summary>
        /// Initialize socket and send out a read request
        /// </summary>
        private void Initialize()
        {
            socket = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);
            ipEndp = new IPEndPoint(IPAddress.Loopback, port);
            endpoint = (EndPoint)ipEndp;
        }

        private void HandleWriteRequest(string filename)
        {
            string path = Path.Combine(FILE_PATH, filename + ".txt");
            using (FileStream fsSource = new FileStream(path,
            FileMode.Open, FileAccess.Read))
            {
                // Read the source file into a byte array.
                int numBytesToRead = (int)fsSource.Length;
                int numBytesRead = 0;
                int packetNumber = 1;

                while (numBytesToRead > 0)
                {
                    byte[] buffer = Create512Buffer((int)PACKET_TYPE.DATA);
                    buffer[2] = (byte)packetNumber;
                    buffer[3] = (byte)(packetNumber >> 8);

                    // Read may return anything from 0 to numBytesToRead.
                    int n = fsSource.Read(buffer, 4, 508);
                    socket.SendTo(buffer, endpoint);
                    // Break when the end of the file is reached.
                    if (n == 0)
                        break;

                    numBytesRead += n;
                    numBytesToRead -= n;
                    bool ackPassed = true;
                    int counter = 0;
                    
                    do
                    {
                        ackPassed = WaitAcknowledge(packetNumber, endpoint);
                        if(!ackPassed)
                        {
                            socket.SendTo(buffer, endpoint);
                        }
                        counter += 1;
                    } while (!ackPassed && counter < 2);
                    if (!ackPassed)
                        break;
                    packetNumber += 1;
                }
                Console.WriteLine($"Sent {numBytesRead} bytes");
            }
        }

        /// <summary>
        /// Waits for client to acknowledge the packet
        /// </summary>
        /// <param name="number"></param>
        /// <param name="remote"></param>
        private static bool WaitAcknowledge(int number, EndPoint remote)
        {
            byte[] buffer = new byte[4];
            socket.ReceiveFrom(buffer, ref remote);

            int opcode = buffer[0] | buffer[1] << 8;
            int block_number = buffer[2] | buffer[3] << 8;

            if (opcode != (int)PACKET_TYPE.ACK || block_number != number)
            {
                SendError(remote, ERROR_CODE.Not_defined, "invalid ack packet");
                return false;
            }
            return true;
        }

        /// <summary>
        /// Sends a error with given error code and message
        /// </summary>
        private static void SendError(EndPoint remote, ERROR_CODE error_code, string message)
        {
            byte[] messageBytes = Encoding.ASCII.GetBytes(message);
            byte[] packet = new byte[4 + messageBytes.Length + 1];
            packet[0] = (byte)PACKET_TYPE.ERROR;
            packet[1] = (byte)((int)PACKET_TYPE.ERROR >> 8);
            packet[2] = (byte)error_code;
            packet[3] = (byte)((int)error_code >> 8);

            int index = 4;
            for (int i = 0; i < messageBytes.Length; i++)
            {
                packet[index] = messageBytes[i];
            }
            socket.SendTo(packet, remote);
        }

        /// <summary>
        /// Creates a data packet 512 bytes 
        /// </summary>
        private static byte[] Create512Buffer(int opcode)
        {
            byte[] buffer = new byte[512];
            buffer[0] = (byte)PACKET_TYPE.DATA;
            buffer[1] = (byte)((int)PACKET_TYPE.DATA >> 8);
            return buffer;
        }

        private void HandleReadRequest()
        {
            int errorCount = 0;
            string path = Path.Combine(FILE_PATH, filename + ".txt");
            using (StreamWriter sw = File.CreateText(path))
            {
                bool conversation = true;
                while (conversation)
                {
                    byte[] buffer = new byte[512];
                    socket.ReceiveFrom(buffer, ref endpoint);
                    int opCode = (int)(buffer[1] << 8 | buffer[0]);
                    switch (opCode)
                    {
                        case (int)PACKET_TYPE.DATA:
                            conversation = ReadData(buffer, sw);
                            SendAck(blockNumber, endpoint);
                            blockNumber += 1;
                            break;
                        default:
                            if(errorCount < 2)
                            {
                                SendAck(blockNumber, endpoint);
                                errorCount += 1;
                            } else if (errorCount >= 2)
                            {
                                conversation = false;
                                Console.WriteLine("Something went wrong...");
                            } else if (opCode == (int)PACKET_TYPE.ERROR)
                            {
                                ReportError(buffer);
                            }
                            break;
                    }
                }
            }
            blockNumber = 1;
        }

        private void SendAck(int number, EndPoint endPoint)
        {
            byte[] buffer = new byte[4];
            buffer[0] = (int)PACKET_TYPE.ACK;
            buffer[1] = (int)PACKET_TYPE.ACK >> 8;
            buffer[2] = (byte)number;
            buffer[3] = (byte)(number >> 8);

            socket.SendTo(buffer, endPoint);
        }

        /// <summary>
        /// Read data from packet and write it to file
        /// </summary>
        /// <returns>true if wasn't last packet (not below 508 bytes)</returns>
        private bool ReadData(byte[] packet, StreamWriter sw)
        {
            string data = String.Empty;

            for(int i = 4; i < packet.Length; i++)
            {
                if (packet[i] == 0) break;
                data += Convert.ToChar(packet[i]);
            }
            sw.Write(data);
            if (data.Length == 508) return true;
            return false;
        } 

        private void ReportError(byte[] packet)
        {
            int errorCode = (int)(packet[3] << 8 | packet[2]);
            switch (errorCode)
            {
                // TODO: Read error message from packet
                case (int)ERROR_CODE.Access_violation:
                    Console.WriteLine("ERROR: Access violation");
                    break;
                case (int)ERROR_CODE.Disk_full:
                    Console.WriteLine("ERROR: Disk full");
                    break;
                case (int)ERROR_CODE.File_already_exists:
                    Console.WriteLine("ERROR: File already exists");
                    break;
                case (int)ERROR_CODE.File_not_found:
                    Console.WriteLine("ERROR: File not found");
                    break;
                case (int)ERROR_CODE.Illegal_TFTP_operation:
                    Console.WriteLine("ERROR: Illegal TFTP operation");
                    break;
                case (int)ERROR_CODE.Not_defined:
                    Console.WriteLine("ERROR: Not defined");
                    break;
                case (int)ERROR_CODE.No_such_user:
                    Console.WriteLine("ERROR: No such user");
                    break;
                case (int)ERROR_CODE.Unknown_transfer_ID:
                    Console.WriteLine("ERROR: Unknown transfer ID");
                    break;
                default:
                    Console.WriteLine("ERROR: Unknown Error code...");
                    break;
            }
        }

        private static void SendError(string error)
        {
            Console.WriteLine(error);
            // TODO: Send error message
        }
    }
}
