﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Text;

namespace TFTPServer
{
    class Server
    {
        public static Socket socket;
        public static IPEndPoint iep;
        public static List<EndPoint> clients;
        public static int port = 69;
        private static int blockNumber = 1;
        private const string FILE_DIR = @"C:\temp";
        private static int socket_timeout = 5000; // milliseconds

        /// <summary>
        /// Enum for packet types
        /// </summary>
        enum PACKET_TYPE
        {
            RRQ = 1,
            WRQ,
            DATA,
            ACK,
            ERROR
        }

        enum ERROR_CODE
        {
            Not_defined,
            File_not_found,
            Access_violation,
            Disk_full,
            Illegal_TFTP_operation,
            Unknown_transfer_ID,
            File_already_exists,
            No_such_user
        }

        /// <summary>
        /// Sends a acknowledgement packet
        /// </summary>
        private static void SendAck(EndPoint remote)
        {
            byte[] packet = new byte[4];
            packet[0] = (byte)PACKET_TYPE.ACK;
            packet[1] = (byte)((int)PACKET_TYPE.ACK >> 8);
            packet[2] = (byte)blockNumber;
            packet[3] = (byte)(blockNumber >> 8);

            socket.SendTo(packet, remote);
        }

        /// <summary>
        /// Sends the file data from path to client
        /// </summary>
        private static void SendData(string path, EndPoint remote)
        {
            using (FileStream fsSource = new FileStream(path,
            FileMode.Open, FileAccess.Read))
            {
                // Read the source file into a byte array.
                int numBytesToRead = (int)fsSource.Length;
                int numBytesRead = 0;
                int packetNumber = 1;

                while (numBytesToRead > 0)
                {
                    byte[] buffer = Create512Buffer((int)PACKET_TYPE.DATA);
                    buffer[2] = (byte)packetNumber;
                    buffer[3] = (byte)(packetNumber >> 8);

                    // Read may return anything from 0 to numBytesToRead.
                    int n = fsSource.Read(buffer, 4, 508);
                    socket.SendTo(buffer, remote);
                    // Break when the end of the file is reached.
                    if (n == 0)
                        break;

                    numBytesRead += n;
                    numBytesToRead -= n;

                    bool ackPassed = true;
                    int counter = 0;

                    do
                    {
                        ackPassed = WaitAcknowledge(packetNumber, remote);
                        if (!ackPassed)
                        {
                            socket.SendTo(buffer, remote);
                        }
                        counter += 1;
                    } while (!ackPassed && counter < 2);
                    if (!ackPassed)
                    {
                        Console.WriteLine("Client didn't respond in time");
                        break;
                    }
                    packetNumber += 1;
                }
                Console.WriteLine($"Sent {numBytesRead} bytes");
            }
        }
        
        /// <summary>
        /// Waits for client to acknowledge the packet
        /// </summary>
        /// <param name="number"></param>
        /// <param name="remote"></param>
        private static bool WaitAcknowledge(int number, EndPoint remote)
        {
            socket.ReceiveTimeout = socket_timeout;
            byte[] buffer = new byte[4];
            try
            {
                socket.ReceiveFrom(buffer, ref remote);

                int opcode = buffer[0] | buffer[1] << 8;
                int block_number = buffer[2] | buffer[3] << 8;

                if (opcode != (int)PACKET_TYPE.ACK || block_number != number)
                {
                    SendError(remote, ERROR_CODE.Not_defined, "invalid ack packet");
                    return false;
                }

                socket.ReceiveTimeout = 30000;
                return true;
            } catch (SocketException)
            {
                socket.ReceiveTimeout = 30000;
                return false;
            }
        }

        /// <summary>
        /// Creates a data packet 512 bytes 
        /// </summary>
        private static byte[] Create512Buffer(int opcode)
        {
            byte[] buffer = new byte[512];
            buffer[0] = (byte)PACKET_TYPE.DATA;
            buffer[1] = (byte)((int)PACKET_TYPE.DATA >> 8);
            return buffer;
        }

        /// <summary>
        /// Sends a error with given error code and message
        /// </summary>
        private static void SendError(EndPoint remote, ERROR_CODE error_code, string message)
        {
            byte[] messageBytes = Encoding.ASCII.GetBytes(message);
            byte[] packet = new byte[4 + messageBytes.Length + 1];
            packet[0] = (byte)PACKET_TYPE.ERROR;
            packet[1] = (byte)((int)PACKET_TYPE.ERROR >> 8);
            packet[2] = (byte)error_code;
            packet[3] = (byte)((int)error_code >> 8);

            int index = 4;
            for(int i = 0; i < messageBytes.Length; i++)
            {
                packet[index] = messageBytes[i];
            }
            socket.SendTo(packet, remote);
        }

        /// <summary>
        /// Get filename from packet
        /// </summary>
        private static string FilenameFromPacket(byte[] packet)
        {
            string filename = String.Empty;
            int index = 2;
            do
            {
                filename += Convert.ToChar(packet[index]);
                index += 1;
            } while (packet[index] != 0);
            return filename;
        }

        /// <summary>
        /// Handles a read request from client
        /// </summary>
        private static void ReadRequest(byte[] buffer, EndPoint remote)
        {
            string filename = FilenameFromPacket(buffer);
            
            Console.WriteLine($"Read request for file : {filename}");

            string file_path = Path.Combine(FILE_DIR, filename + ".txt");
            if (!String.IsNullOrEmpty(filename) && File.Exists(file_path))
            {
                SendData(file_path, remote);
            }
            else
            {
                string message = "File not found.";
                SendError(remote, ERROR_CODE.File_not_found, message);
            }
        }

        /// <summary>
        /// Handles a write request from client
        /// </summary>
        private static void WriteRequest(byte[] buffer, EndPoint remote)
        {
            string filename = FilenameFromPacket(buffer);
            string path     = Path.Combine(FILE_DIR, filename + ".txt");
            
            if(File.Exists(path)) {
                string message = "File already exists";
                SendError(remote, ERROR_CODE.File_already_exists, message);
                return;
            }

            SendAck(remote);
            HandleWRQ(remote, filename);
        }

        /// <summary>
        /// Listens clients data packets
        /// </summary>
        private static void HandleWRQ(EndPoint remote, string filename)
        {
            string path = Path.Combine(FILE_DIR, filename + ".txt");    
            bool conversation = true;
            int errorCount = 0;

            // Create a file to write to.
            using (StreamWriter sw = File.CreateText(path))
            {
                while (conversation)
                {
                    byte[] packet = new byte[512];
                    socket.ReceiveFrom(packet, ref remote);
                    int opcode = packet[0] | packet[1] << 8;
                    int blockNumber = packet[2] | packet[3] << 8;

                    if (opcode == (int)PACKET_TYPE.DATA && blockNumber == (blockNumber + 1))
                    {
                        conversation = ReadData(packet, sw);
                        SendAck(remote);
                        blockNumber += 1;
                    } else if (errorCount < 1)
                    {
                        SendAck(remote);
                        errorCount += 1;
                    }
                    else if (opcode == (int)PACKET_TYPE.ERROR)
                    {
                        ErrorResponse(packet);
                        conversation = false;
                    }
                }
            }
        }

        /// <summary>
        /// Read data from packet and write it to file
        /// </summary>
        /// <returns>true if wasn't last packet (not below 508 bytes)</returns>
        private static bool ReadData(byte[] packet, StreamWriter sw)
        {
            string data = String.Empty;

            for (int i = 4; i < packet.Length; i++)
            {
                if (packet[i] == 0) break;
                data += Convert.ToChar(packet[i]);
            }
            sw.Write(data);
            if (data.Length == 508) return true;
            return false;
        }

        private static void ErrorResponse(byte[] buffer)
        {
            int errorCode = (int)(buffer[3] << 8 | buffer[2]);
            switch (errorCode)
            {
                // TODO: Read error message from packet
                case (int)ERROR_CODE.Access_violation:
                    Console.WriteLine("ERROR: Access violation");
                    break;
                case (int)ERROR_CODE.Disk_full:
                    Console.WriteLine("ERROR: Disk full");
                    break;
                case (int)ERROR_CODE.File_already_exists:
                    Console.WriteLine("ERROR: File already exists");
                    break;
                case (int)ERROR_CODE.File_not_found:
                    Console.WriteLine("ERROR: File not found");
                    break;
                case (int)ERROR_CODE.Illegal_TFTP_operation:
                    Console.WriteLine("ERROR: Illegal TFTP operation");
                    break;
                case (int)ERROR_CODE.Not_defined:
                    Console.WriteLine("ERROR: Not defined");
                    break;
                case (int)ERROR_CODE.No_such_user:
                    Console.WriteLine("ERROR: No such user");
                    break;
                case (int)ERROR_CODE.Unknown_transfer_ID:
                    Console.WriteLine("ERROR: Unknown transfer ID");
                    break;
                default:
                    Console.WriteLine("ERROR: Unknown Error code...");
                    break;
            }
        }

        private static void AckResponse(byte[] buffer)
        {
            throw new NotImplementedException();
        }

        private static void DataRequest(byte[] buffer)
        {
            throw new NotImplementedException();
        }

        private static void Initialize()
        {
            clients = new List<EndPoint>();
            iep = new IPEndPoint(IPAddress.Loopback, port);
            try
            {
                socket = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);
                socket.Bind(iep);
            }
            catch (Exception ex)
            {
                Console.WriteLine("Error... " + ex.Message);
                Console.ReadKey();
                return;
            }

            Console.WriteLine("TFTP server up, listening for clients...");

            while (!Console.KeyAvailable)
            {
                byte[] buffer = new byte[256];
                IPEndPoint asiakas = new IPEndPoint(IPAddress.Any, 0);
                EndPoint remote = (EndPoint)asiakas;
                socket.ReceiveFrom(buffer, ref remote);

                int opcode = buffer[0] + buffer[1];
                switch (opcode) {
                    case (int)PACKET_TYPE.RRQ:
                        ReadRequest(buffer, remote);
                        break;
                    case (int)PACKET_TYPE.WRQ:
                        WriteRequest(buffer, remote);
                        break;
                    default:
                        break;
                }
            }

            Console.ReadKey();
            socket.Close();
        }

        static void Main(string[] args)
        {
            Initialize();
        }
    }
}


